Namespace MultiBandit
  Friend Class MultiBanditProgram
	Shared Sub Main(ByVal args() As String)
            Console.WriteLine(vbCrLf & "Begin multi-armed bandit problem demo " & vbCrLf)

            Console.WriteLine("Creating 3 Gaussian machines")
            Console.WriteLine("Machine 0 mean =  0.0, sd = 1.0")
            Console.WriteLine("Machine 1 mean = -0.5, sd = 2.0")
            Console.WriteLine("Machine 2 mean =  0.1, sd = 0.5")
            Console.WriteLine(vbCrLf & "Best machine is [2] with mean pay = 0.1")

            Dim nMachines As Integer = 3
            Dim machines(nMachines - 1) As Machine
            machines(0) = New Machine(0.0, 1.0, 0)
            machines(1) = New Machine(-0.5, 2.0, 1)
            machines(2) = New Machine(0.1, 0.5, 2)

            Dim nPulls As Integer = 20
            Dim pctExplore As Double = 0.4
            Console.WriteLine("Setting nPulls = " & nPulls)

            Console.WriteLine(vbCrLf & "Using explore-exploit with pctExplore = " & pctExplore.ToString("F2"))
            Dim avgPay As Double = ExploreExploit(machines, pctExplore, nPulls)
            Dim totPay As Double = avgPay * nPulls
            Console.WriteLine(vbCrLf & "Average pay per pull = " & avgPay.ToString("F2"))
            Console.WriteLine("Total payout         = " & totPay.ToString("F2"))

            Dim avgBase As Double = machines(2).mean
            Dim totBase As Double = avgBase * nPulls
            Console.WriteLine(vbCrLf & "Baseline average pay = " & avgBase.ToString("F2"))
            Console.WriteLine("Total baseline pay   = " & totBase.ToString("F2"))

            Dim regret As Double = totBase - totPay
            Console.WriteLine(vbCrLf & "Total regret = " & regret.ToString("F2"))

            Console.WriteLine(vbCrLf & "End bandit demo " & vbCrLf)
            Console.ReadLine()
        End Sub ' Main

        Private Shared Function ExploreExploit(ByVal machines() As Machine, ByVal pctExplore As Double, ByVal nPulls As Integer) As Double
            ' use basic explore-exploit algorithm
            ' return the average pay per pull
            Dim nMachines As Integer = machines.Length
            Dim r As New Random(2) ' which machine (2 gives representative demo)

            Dim explorePays(nMachines - 1) As Double ' curr total payout (init 0.0 all)
            Dim totPay As Double = 0.0

            Dim nExplore As Integer = CInt(Fix(nPulls * pctExplore)) ' n pulls to explore
            Dim nExploit As Integer = nPulls - nExplore ' n pulls to exploit

            Console.WriteLine(vbCrLf & "Start explore phase")
            For pull = 0 To nExplore - 1 ' explore phase
                Dim m As Integer = r.Next(0, nMachines) ' pick a machine at random
                Dim pay As Double = machines(m).Pay ' play
                Console.Write("[" & pull.ToString.PadLeft(3) & "]  ")
                Console.WriteLine("selected machine " & m & ". pay = " & pay.ToString("F2").PadLeft(6))
                explorePays(m) += pay ' update
                totPay += pay
            Next pull ' explore

            Dim bestMach As Integer = BestIdx(explorePays) ' which machine paid most
            Console.WriteLine(vbCrLf & "Best machine found during explore = " & bestMach)

            Console.WriteLine(vbCrLf & "Start exploit phase")
            For pull = 0 To nExploit - 1 ' exploit phase
                Dim pay As Double = machines(bestMach).Pay ' play same machine always
                Console.Write("[" & pull.ToString.PadLeft(3) & "]  ")
                Console.WriteLine("pay = " & pay.ToString("F2").PadLeft(6))
                totPay += pay ' accumulate
            Next pull ' exploit

            Return totPay / nPulls ' avg payout per pull
        End Function ' ExploreExploit

        Private Shared Function BestIdx(ByVal pays() As Double) As Integer
            ' index of array with largest value
            Dim result As Integer = 0
            Dim maxVal As Double = pays(0)
            For i = 0 To pays.Length - 1
                If pays(i) > maxVal Then
                    result = i
                    maxVal = pays(i)
                End If
            Next i
            Return result
        End Function

        'Shared Sub Display(ByVal array() As Integer)
        '    For i As Integer = 0 To array.Length - 1
        '        Console.Write(arrayi & " ")
        '    Next i
        '    Console.WriteLine("")
        'End Sub

        'Shared Sub Display(ByVal array() As Double)
        '    For i As Integer = 0 To array.Length - 1
        '        Console.Write(arrayi.ToString("F4") & " ")
        '    Next i
        '    Console.WriteLine("")
        'End Sub

    End Class ' Program class

    Public Class Machine
        Public mean As Double ' avg payout per pull
        Public sd As Double ' variability about the mean
        Private g As Gaussian ' payout generator

        Public Sub New(ByVal mean As Double, ByVal sd As Double, ByVal seed As Integer)
            Me.mean = mean
            Me.sd = sd
            Me.g = New Gaussian(mean, sd, seed)
        End Sub

        Public Function Pay() As Double
            Return Me.g.Next
        End Function

        ' -----

        Private Class Gaussian
            Private r As Random
            Private mean As Double
            Private sd As Double

            Public Sub New(ByVal mean As Double, ByVal sd As Double, ByVal seed As Integer)
                Me.r = New Random(seed)
                Me.mean = mean
                Me.sd = sd
            End Sub

            Public Function [Next]() As Double
                Dim u1 As Double = r.NextDouble
                Dim u2 As Double = r.NextDouble
                Dim left As Double = Math.Cos(2.0 * Math.PI * u1)
                Dim right As Double = Math.Sqrt(-2.0 * Math.Log(u2))
                Dim z As Double = left * right
                Return Me.mean + (z * Me.sd)
            End Function
        End Class

        ' -----
    End Class ' Machine
End Namespace ' ns
